/*
	d80211 Marvell Libertas PCI driver

	Copyright (c) 2007 Andrea Merello <andreamrl@tiscali.it>
	
	Based on existent mrv8k linux alpha driver
	Copyright (c) 2005 Luc Saillard <luc@saillard.org>
	
	Based on BSD malo driver	
	Copyright (c) 2006 Claudio Jeker <claudio@openbsd.org>
 	Copyright (c) 2006 Marcus Glocker <mglocker@openbsd.org>
	(no code has been copyed. Hw programming methods
	magic values, and card specific programming details
	had been looked at)

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; see the file COPYING.  If not, write to
	the Free Software Foundation, Inc., 51 Franklin Steet, Fifth Floor,
	Boston, MA 02110-1301, USA.

*/

#include <linux/firmware.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/etherdevice.h>
#include <asm/semaphore.h>
#include <net/mac80211.h>

#include "mrv8.h"

MODULE_AUTHOR(DRV_COPYRIGHT);
MODULE_DESCRIPTION (DRV_DESCRIPTION);
MODULE_VERSION(DRV_VERSION);
MODULE_LICENSE("GPL");

static struct pci_device_id mrv_pci_tbl[] = {
	{PCI_VENDOR_ID_MARVELL, PCI_DEVICE_ID_MARVELL_W8K, 0x1FA4, 0x11AB, 0, 0, WL3563 },
	{PCI_VENDOR_ID_MARVELL, PCI_DEVICE_ID_MARVELL_W8K, 0x1FA5, 0x16AB, 0, 0, WL3563 },
	{PCI_VENDOR_ID_MARVELL, PCI_DEVICE_ID_MARVELL_W8K, 0x1FA6, 0x16AB, 0, 0, WL3563 },
	{PCI_VENDOR_ID_MARVELL, PCI_DEVICE_ID_MARVELL_W8K, 0x1FA7, 0x16AB, 0, 0, WL3563 },
	{PCI_VENDOR_ID_MARVELL, PCI_DEVICE_ID_MARVELL_W8K, 0x138f, 0x1043, 0, 0, WL138G },
	{PCI_VENDOR_ID_MARVELL, PCI_DEVICE_ID_MARVELL_W8K, 0x128f, 0x1043, 0, 0, WL138G },
	{PCI_VENDOR_ID_MARVELL, PCI_DEVICE_ID_MARVELL_W8K, 0x108f, 0x1043, 0, 0, WL138G },
		/* NETGEAR WG311v3 802.11g Wireless PCI Adapter */
	{PCI_VENDOR_ID_MARVELL, 0x1faa, PCI_ANY_ID, PCI_ANY_ID, 0x6b00, 0x1385, W8335},
	{PCI_VENDOR_ID_MARVELL, 0x1faa, PCI_ANY_ID, PCI_ANY_ID, 0, 0, W8335},
	{PCI_VENDOR_ID_MARVELL, PCI_DEVICE_ID_MARVELL_W8K, PCI_ANY_ID, PCI_ANY_ID, 0, 0, WLGENERIC },
	{0,}
};

static void mrv_deinit_cmd(struct mrv_priv *priv)
{
	if(priv->cmd_virt)
		pci_free_consistent(priv->pdev,
			MRV_CMD_PKT_SIZE,
			priv->cmd_virt, priv->cmd_dma);
	
	if(priv->magic_virt)
		pci_free_consistent(priv->pdev, 4,
			priv->magic_virt, priv->magic_dma);
}

static void mrv_free_rx_ring(struct mrv_priv *priv)
{
	int i;
	for(i=0;i<MRV_RX_DESC_NUM;i++){
		dev_kfree_skb(priv->rx_skb[i]);
		pci_free_consistent(priv->pdev,	
			sizeof(struct mrv_rx_descriptor),
			priv->rx_ring_virt[i], 
			priv->rx_ring_dma[i]);
	}

}

static void mrv_free_tx_ring(struct mrv_priv *priv)
{
	int i,j;
	for(i=0;i<MRV_TX_DESC_NUM;i++)
		pci_free_consistent(priv->pdev,
			sizeof(struct mrv_tx_descriptor),
			priv->tx_ring_virt[i],
			priv->tx_ring_dma[i]);
	
	for(i=0;i<priv->tx_pending;i++){
		j=priv->tx_r_idx;
		dev_kfree_skb(priv->tx_skb[j]);
		if(j == MRV_TX_DESC_NUM - 1)
			j=0;
		else	
			j++;
	}
}

static int mrv_init_cmd(struct mrv_priv *priv)
{	
	void *v;
	dma_addr_t p;
	v = pci_alloc_consistent(priv->pdev, MRV_CMD_PKT_SIZE, &p);
	priv->cmd_virt = v;
	
	if( v == NULL){
		printk(KERN_ERR"Can't allocate CMD memory\n");
		return -ENOMEM;
	}
	
	priv->cmd_dma = p;
	v = pci_alloc_consistent(priv->pdev, 4, &p);
	priv->magic_virt = v;
	
	if( v == NULL){
		printk(KERN_ERR"Can't allocate MAGIC memory\n");
		return -ENOMEM;
	}
	
	priv->magic_dma = p;
	*priv->magic_virt = cpu_to_le32(MRV_CMD_MAGIC);

	return 0;
}

static void _mrv_send_cmd(struct mrv_priv *priv, u32 addr)
{
	wmb();
	mrv_reg_write32(priv, MRV_REG_CMD_PTR,  addr);
	mrv_reg_mb(priv);
	mrv_reg_write32(priv, MRV_REG_GEN_ARM_INT, 2); 
	mrv_reg_mb(priv);
}

static inline void mrv_send_cmd(struct mrv_priv *priv)
{
	_mrv_send_cmd(priv, priv->cmd_dma);
}

static int mrv_send_cmd_sync(struct mrv_priv *priv)
{
	int err;

	down(&priv->cmd_sem);
	err = 0;
	priv->cmd_status = 0x10000;
	mrv_send_cmd(priv);

	wait_event_interruptible_timeout(priv->wait_command_queue,
					     (priv->cmd_status != 0x10000), 500);

	if(priv->cmd_status == 0x10000){
		printk(KERN_ERR"Command timeout\n");
		err= -1;
	}else if(priv->cmd_status == 3){
		printk(KERN_WARNING"Command deferred!\n");
		err = 0;
	 }else if(priv->cmd_status != 0){
		printk(KERN_ERR"Command failed with status %x\n", priv->cmd_status);
		err= -1;
	}
	
	up(&priv->cmd_sem);

	return err;
}

static inline struct mrv_cmd_hdr *mrv_cmd(struct mrv_priv *priv, int cmd, int len, int status, int sq)
{
	struct mrv_cmd_hdr *ret;
	
	ret=(struct mrv_cmd_hdr *)priv->cmd_virt;
	ret->type = cpu_to_le16(cmd);
	ret->len = cpu_to_le16(len);
	ret->status = cpu_to_le16(status);
	ret->sq = cpu_to_le16(sq);
	return ret;
}

static int mrv_upload_fw(struct mrv_priv *priv)
{
	struct mrv_cmd_hdr  *cmd;
	u8 *cmd_data;
	const struct firmware *img = NULL;
	int i,j,sq,err;

	err = request_firmware(&img, FW_BOOT_FILENAME, &priv->pdev->dev);
	if (err < 0) {
		printk(KERN_ERR"request_firmware for boot img failed: Reason %d\n", err);
		return err;
	}
	
	mrv_reg_write32(priv, MRV_REG_STATUS, 0);
	mrv_reg_mb(priv);
	mrv_mem_write16(priv, MRV_MEM_CMD_BOOT, MRV_CMD_FW);
	mrv_mem_write16(priv, MRV_MEM_CMD_BOOT+2, img->size);
	mrv_mem_write32(priv, MRV_MEM_CMD_BOOT+4, 0);
		
	for(i=0;i< img->size;i++)
		mrv_mem_write8(priv, MRV_MEM_CMD_BOOT+8+i, (img->data)[i]);
	
	release_firmware(img);
	
	_mrv_send_cmd(priv, MRV_HWMEM_CMD_BOOT);

	for (i = 0; i < 500; i++) {
		if (mrv_reg_read32(priv, MRV_REG_STATUS) == 0x5)
		mdelay(1);
			break;
	}
	if (i == 10) {
		printk(KERN_ERR"Timeout while loading boot fw \n");
		return -1;
	}
	printk(KERN_INFO"Bootstrap loaded OK, jumping in..\n");
	mrv_mem_write16(priv, MRV_MEM_CMD_BOOT, MRV_CMD_FW);
	mrv_mem_write16(priv, MRV_MEM_CMD_BOOT+2, 0);
	mrv_mem_write32(priv, MRV_MEM_CMD_BOOT+4, 0);
	
	_mrv_send_cmd(priv, MRV_HWMEM_CMD_BOOT);

	/* the boot code is stupid enought that it does not report any status when it
	 * loads correctly and become alive. Wait 1 second..*/
	msleep(1000);
	
	err = request_firmware(&img, FW_FIRMWARE_FILENAME, &priv->pdev->dev);
	if (err < 0) {
		printk(KERN_ERR"request_firmware for fw img2 failed: Reason %d\n", err);
		return err;
	}

	sq=1;
	for(i=0; (i+256) < img->size; i+=256){
		mrv_reg_write32(priv, MRV_REG_STATUS, 0);
		mrv_reg_mb(priv);
		cmd = mrv_cmd(priv, MRV_CMD_FW, 256, 0, sq++);
		cmd_data = MRV_CMD_PAYLOAD(cmd) ;
		memcpy(cmd_data, img->data + i, 256);
		mrv_send_cmd(priv);
		for(j=0;j<500;j++){
			if(mrv_reg_read32(priv, MRV_REG_STATUS) == 5) break;
			mdelay(1);
		}
		if(j==500){
			printk(KERN_ERR"Boot FW has failed to transfer real FW\n");
			return -2;
		}
	}

	if(i < img->size){
		mrv_reg_write32(priv, MRV_REG_STATUS, 0);
		mrv_reg_mb(priv);
		cmd = mrv_cmd(priv, MRV_CMD_FW, img->size - i, 0, sq++);
		cmd_data = MRV_CMD_PAYLOAD(cmd) ;
		memcpy(cmd_data, img->data + i, img->size - i);
		mrv_send_cmd(priv);
		for(j=0;j<500;j++){
			if(mrv_reg_read32(priv, MRV_REG_STATUS) == 5) break;
			mdelay(1);
		}
		if(j==500){
			printk(KERN_ERR"Boot FW has failed to transfer last real FW chunk\n");
			return -2;
		}
	}

	release_firmware(img);
	
	cmd = mrv_cmd(priv, MRV_CMD_FW, 0, 0, sq);
	
	mrv_send_cmd(priv);
	msleep(100);
	
	for( i=0; i<200; i++){
		mrv_reg_write32(priv, MRV_REG_CMD_PTR, 0x5a);
		mrv_reg_mb(priv);
		if(mrv_reg_read32(priv, MRV_REG_STATUS) == MRV_FW_SIGNATURE)
			break;
		mdelay(10);
	}	
	
	if(i == 200){
		printk(KERN_ERR"FW is not alive or returned error %x \n",
			mrv_reg_read32(priv,MRV_REG_STATUS));
		return -1;
	}

	priv->fw_load = 1;
	printk(KERN_INFO"Firmware load successfully\n");

	return 0;

}

static void mrv_int_enable(struct mrv_priv *priv)
{
	mrv_reg_write32(priv, MRV_REG_INTA_MASK, MRV_INTA_MASK);
	mrv_reg_write32(priv, MRV_REG_INTB_MASK, MRV_INTB_MASK);
	mrv_reg_mb(priv);
}

static void mrv_int_disable(struct mrv_priv *priv)
{
	mrv_reg_write32(priv, MRV_REG_INTA_MASK, 0);
	mrv_reg_write32(priv, MRV_REG_INTB_MASK, 0);
	mrv_reg_mb(priv);
}

static int mrv_tx(struct ieee80211_hw *dev, struct sk_buff *skb, struct ieee80211_tx_control *control)
{
	struct ieee80211_hdr *hdr;
	struct mrv_tx_descriptor *txd;
	unsigned long flags;
	struct mrv_priv *priv = dev->priv;
	
	spin_lock_irqsave(&priv->tx_lock, flags);
	
	if(skb->len < sizeof(struct ieee80211_hdr)){
		spin_unlock_irqrestore(&priv->tx_lock, flags);
		return NETDEV_TX_OK;
	}

	if(priv->tx_pending >= (MRV_TX_DESC_NUM -2)){
		ieee80211_stop_queue(dev,0);
	}

	if (skb_headroom(skb) < dev->extra_tx_headroom) {
		if (pskb_expand_head(skb, dev->extra_tx_headroom, 0, GFP_ATOMIC)) {
			printk(KERN_ERR "failed to allocate room for TX "
			       "header\n");
			kfree_skb(skb);
			spin_unlock_irqrestore(&priv->tx_lock, flags);
			return -ENOMEM;
		}
	}

	priv->tx_pending++;
	priv->tx_skb[priv->tx_w_idx] = skb;
	txd = priv->tx_ring_virt[priv->tx_w_idx];
	
	if(priv->tx_w_idx == MRV_TX_DESC_NUM -1)
		priv->tx_w_idx = 0;
	else
		priv->tx_w_idx++;

	memcpy(txd->dest_addr,
		ieee80211_get_DA((struct ieee80211_hdr *)skb->data), 6);

	skb_push(skb, dev->extra_tx_headroom);

	*((u16*)skb->data) = cpu_to_le16(skb->len) -32;

	memcpy(skb->data+2, skb->data+8, 24); /* move 80211 header */ 
	hdr = (struct ieee80211_hdr *)skb->data + 2;
	memcpy(skb->data+2+24, &(hdr->addr4), 6); /* copy addr4 */

	txd->len = cpu_to_le16(skb->len);
	
	txd->dma_addr = cpu_to_le32(pci_map_single(priv->pdev,
			skb->data, skb->len, PCI_DMA_TODEVICE));
	
	txd->priority = control->queue;
	txd->rate = control->tx_rate;

	txd->status = cpu_to_le16(1);
	txd->fwowner = cpu_to_le16(0x8000);
	wmb();
	mrv_reg_write32(priv, MRV_REG_GEN_ARM_INT, 1);
	mrv_reg_mb(priv);

	spin_unlock_irqrestore(&priv->tx_lock, flags);

	return NETDEV_TX_OK;
}

static void mrv_tx_cb_tasklet(struct mrv_priv *priv)
{
	struct mrv_tx_descriptor *txd;
	int status;
	unsigned long flags;
	struct ieee80211_tx_status tx_status = {{0}};

	spin_lock_irqsave(&priv->tx_lock, flags);

	while(priv->tx_r_idx != priv->tx_w_idx){
		txd = priv->tx_ring_virt[priv->tx_r_idx];
		
		if(le16_to_cpu(txd->fwowner) & 0x8000)
			break;
		
		status = le16_to_cpu(txd->status) & 0x1;
		if(status == 0) tx_status.excessive_retries = 1;
		tx_status.queue_number = txd->priority;

		pci_unmap_single(priv->pdev, txd->dma_addr,
			priv->tx_skb[priv->tx_r_idx]->len, PCI_DMA_TODEVICE); 
			
		ieee80211_tx_status_irqsafe(priv->dev, 
			priv->tx_skb[priv->tx_r_idx], &tx_status);
	
		priv->tx_pending--;

		if(priv->tx_pending == MRV_TX_DESC_NUM -2)
			ieee80211_wake_queue(priv->dev,0);
	
		if(priv->tx_r_idx == MRV_TX_DESC_NUM -1)
			priv->tx_r_idx = 0;
		else
			priv->tx_r_idx++;
	}

	spin_unlock_irqrestore(&priv->tx_lock, flags);
}

static void mrv_rx_tasklet(struct mrv_priv *priv)
{
	int len, count = 0;
	struct ieee80211_rx_status rx_status = {0};
	struct mrv_rx_descriptor *rxd;
	struct sk_buff *skb;

	for(count = 0;count < MRV_MAX_RX_INT; count++){
		rxd = priv->rx_ring_virt[priv->rx_idx];
		skb = priv->rx_skb[priv->rx_idx];

		if((rxd->drvowner & 0x80)==0)
			break;
		if(mrv_mem_read32(priv, priv->RxHwDmaR) ==
			mrv_mem_read32(priv, priv->RxHwDmaW))
			break;
		
		len = le16_to_cpu(rxd->datalen);
		
		if(len > 3){ 

			rx_status.ssi = rxd->rssi;		
	//		rx_status.antenna = (hdr->signal >> 7) & 1;
	//		rx_status.signal = quality;
			rx_status.rate = rxd->rate;
			rx_status.freq = mrv_channels[rxd->chan -1].freq;
			rx_status.channel = rxd->chan;
			rx_status.phymode = priv->dev->conf.phymode;
	//		rx_status.mactime = le64_to_cpu(hdr->mac_time);
	
			pci_unmap_single(priv->pdev, rxd->dma_data,
				MRV_RX_PAYLOAD_LEN, PCI_DMA_FROMDEVICE); 
	
			/*This is the madness of how firmware reworks packets...
			* So we have to cut out this damn garbage the firmware
			* puts at the beginning and in the middle of packets.....
			*/
			skb_put(skb, len);
			skb_pull(skb, 2);
		
			if(skb->len > 30){
				memcpy(skb->data + 24, skb->data + 30, skb->len - 6);
				skb_trim(skb, skb->len - 6);
			}
//if(skb->len < 50){
//printk("\n.........");
//for(i=0;i<skb->len;i++)
//printk("%x ", skb->data[i]);
//}
			ieee80211_rx_irqsafe(priv->dev, skb, &rx_status);
			
			/*TODO check for failure */
			skb = dev_alloc_skb(MRV_RX_PAYLOAD_LEN);
			priv->rx_skb[priv->rx_idx] = skb;

			rxd->dma_data = cpu_to_le32(pci_map_single(priv->pdev,
					skb->data, MRV_RX_PAYLOAD_LEN, PCI_DMA_FROMDEVICE));
		}

		rxd->drvowner = 0;
		
		if(priv->rx_idx == MRV_RX_DESC_NUM -1)
			priv->rx_idx = 0;
		else
			priv->rx_idx++;
		wmb();
		mrv_mem_write32(priv, priv->RxHwDmaR, priv->rx_ring_dma[priv->rx_idx]);
	}

	//rxd = priv->rx_ring_virt[priv->rx_idx];
	//if(priv->up && ((rxd->drvowner & 0x80)!= 0))
	//	tasklet_schedule(&priv->mrv_rx_tasklet);
}

irqreturn_t mrv_interrupt(int irq, void *data)
{
	struct mrv_priv *priv = (struct mrv_priv *) data;
	struct mrv_cmd_hdr *cmd;
	unsigned int status;
	unsigned int flags;
	
	spin_lock_irqsave(&priv->irq_th_lock, flags);
	
	if(!priv->fw_load){
		spin_unlock_irqrestore(&priv->irq_th_lock, flags);
		return IRQ_NONE;
	}

	mrv_int_disable(priv);

	status = mrv_reg_read32(priv, MRV_REG_ISR);
	
	if (status == 0xFFFFFFFF){
		spin_unlock_irqrestore(&priv->irq_th_lock, flags);
		mrv_int_enable(priv);
		return IRQ_NONE;

	}
	status &= MRV_INTA_MASK;
	
	if (status == 0){
		spin_unlock_irqrestore(&priv->irq_th_lock, flags);
		mrv_int_enable(priv);
		return IRQ_NONE;
	}

	if (status & 0x4){
		cmd=(struct mrv_cmd_hdr *)priv->cmd_virt;	
		priv->cmd_status = le16_to_cpu(cmd->status);
		wake_up_interruptible(&priv->wait_command_queue);
	}

	if (status & 0x2)
		tasklet_schedule(&priv->mrv_rx_tasklet);

	if (status & 0x1)
		tasklet_schedule(&priv->mrv_tx_cb_tasklet);

	status = status &~ (0x4|0x2|0x1);

	mrv_reg_write32(priv, MRV_REG_ISR, status);
	mrv_reg_mb(priv);

	mrv_int_enable(priv);
	spin_unlock_irqrestore(&priv->irq_th_lock, flags);
	return IRQ_HANDLED;
}

static void dump_reg(struct mrv_priv *priv)
{
	/*
	 0 HIU host interf 0c pci interf
	 2 CIU
	 4 SQRAM
	 6 SDRAM
	 8 FEMAC
	 a WLMAC. a0 WLMACTXCONFIG ,a2 WLMACRXCONFIG
	 c LBU. c2 eeprom, c4 rf, c6 BB, c8 lsbu
	 d GPIO. d8, pwrmgmt
	 e DMA
	*/
	int i; 
	for(i=0;i<0xfff;i++){
		if(i% 18==0)printk("\n");
		printk("%4x= %2x, ",i,mrv_reg_read8(priv,i));
	}
}

static int mrv_hw_start(struct mrv_priv *priv)
{
	int err;

	err = request_irq(priv->pdev->irq, &mrv_interrupt, IRQF_SHARED,
		          DRV_NAME, priv);
	if (err) {
		printk(KERN_ERR"Error calling request_irq: %d.\n", priv->pdev->irq);
		return err;
	}

	mrv_reg_read32(priv, MRV_REG_ISR);
	mrv_reg_write32(priv, MRV_REG_ISR, 0);
	mrv_reg_write32(priv, 0xc20,0x800f);
	mrv_reg_write32(priv, 0xc24,0x800f);
	mrv_int_disable(priv);
	
	err = mrv_upload_fw(priv);

	if(err){
		free_irq(priv->pdev->irq, priv);
		return err;
	}

	mrv_int_enable(priv);
	
	return 0;
}

static void mrv_hw_halt(struct mrv_priv *priv)
{
	mrv_reg_write32(priv, MRV_REG_GEN_ARM_INT, (1 << 15));
	mrv_reg_mb(priv);
	mrv_int_disable(priv);
	free_irq(priv->pdev->irq, priv);
}

static int mrv_set_antenna(struct mrv_priv *priv, int rtx, int param)
{
	struct mrv_cmd_hdr  *cmd;
	struct mrv_cmd_set_antenna *data;
	
	cmd = mrv_cmd(priv, MRV_CMD_SET_ANTENNA,
		MRV_CMD_SET_ANTENNA_SZ, 0 ,1);
	
	data = (struct mrv_cmd_set_antenna *)MRV_CMD_PAYLOAD(cmd);

	data->rtx = cpu_to_le16(rtx);
	data->param = cpu_to_le16(param);

	if(mrv_send_cmd_sync(priv)){
		printk(KERN_ERR"Could not set antenna\n");
		return -1;
	}

	return 0;
}

static int mrv_set_txpower(struct mrv_priv *priv, int param)
{
	struct mrv_cmd_hdr  *cmd;
	struct mrv_cmd_set_txpower *data;
	
	cmd = mrv_cmd(priv, MRV_CMD_SET_TXPOWER,
		MRV_CMD_SET_TXPOWER_SZ, 0 ,1);
	
	data = (struct mrv_cmd_set_txpower *)MRV_CMD_PAYLOAD(cmd);

	data->txpower = cpu_to_le16(param);

	if(mrv_send_cmd_sync(priv)){
		printk(KERN_ERR"Could not set txpower\n");
		return -1;
	}

	return 0;
}

static int mrv_set_radio(struct mrv_priv *priv, int enable, int preamble)
{
	struct mrv_cmd_hdr  *cmd;
	struct mrv_cmd_set_radio *data;
	
	cmd = mrv_cmd(priv, MRV_CMD_SET_RADIO,
		MRV_CMD_SET_RADIO_SZ, 0 ,1);
	
	data = (struct mrv_cmd_set_radio *)MRV_CMD_PAYLOAD(cmd);

	data->flag = cpu_to_le16(1);
	data->preamble = cpu_to_le16(preamble); 
	data->enable = cpu_to_le16(enable);

	if(mrv_send_cmd_sync(priv)){
		printk(KERN_ERR"Could not set radio\n");
		return -1;
	}

	return 0;
}

static int mrv_set_allowed_rates(struct mrv_priv *priv, int allowg)
{
	struct mrv_cmd_hdr  *cmd;
	struct mrv_cmd_set_rates *data;
	int i;

	cmd = mrv_cmd(priv, MRV_CMD_SET_RATES,
		MRV_CMD_SET_RATES_SZ, 0 ,1);
	
	data = (struct mrv_cmd_set_rates *)MRV_CMD_PAYLOAD(cmd);

	data->fixedrate = 0;
	data->fixedindex = 0;

	for(i=0;i<4;i++)
		data->rates[i] = i;
	
	for(;i<14;i++)
		if(allowg) data->rates[i] = i;
		else data->rates[i] = 0;

	if(mrv_send_cmd_sync(priv)){
		printk(KERN_ERR"Could not set rates\n");
		return -1;
	}
		
	return 0;
}

static int mrv_set_slot(struct mrv_priv *priv, int shortslot)
{
	struct mrv_cmd_hdr  *cmd;
	struct mrv_cmd_set_slot *data;
	
	cmd = mrv_cmd(priv, MRV_CMD_SET_SLOT,
		MRV_CMD_SET_SLOT_SZ, 0 ,1);
	
	data = (struct mrv_cmd_set_slot *)MRV_CMD_PAYLOAD(cmd);

	data->action = cpu_to_le16(1);
	data->shortslot = shortslot; 

	if(mrv_send_cmd_sync(priv)){
		printk(KERN_ERR"Could not set slot\n");
		return -1;
	}
	
	return 0;
}

static int mrv_set_channel(struct mrv_priv *priv, int ch)
{
	struct mrv_cmd_hdr  *cmd;
	struct mrv_cmd_set_channel *data;
	
	cmd = mrv_cmd(priv, MRV_CMD_SET_CHANNEL,
		MRV_CMD_SET_CHANNEL_SZ, 0 ,1);
	
	data = (struct mrv_cmd_set_channel *)MRV_CMD_PAYLOAD(cmd);

	data->flag = cpu_to_le16(1);
	data->ch = ch; 

	if(mrv_send_cmd_sync(priv)){
		printk(KERN_ERR"Could not set channel\n");
		return -1;
	}
	
	return 0;
}

static int mrv_set_mac_addr(struct mrv_priv *priv, u8 addr[6])
{
	struct mrv_cmd_hdr  *cmd;
	struct mrv_cmd_set_mac_addr *data;
	
	cmd = mrv_cmd(priv, MRV_CMD_SET_MAC_ADDR,
		MRV_CMD_SET_MAC_ADDR_SZ, 0 ,1);
	
	data = (struct mrv_cmd_set_mac_addr *)MRV_CMD_PAYLOAD(cmd);

	memcpy(data->addr, addr, 6);

	if(mrv_send_cmd_sync(priv)){
		printk(KERN_ERR"Could not set mac address\n");
		return -1;
	}
	
	return 0;
}

static void mrv_set_bssid(struct mrv_priv *priv, u8 addr[6])
{
	int i;
	u32 reg;

	reg = mrv_reg_read32(priv, MRV_REG_RX_MODE);
	
	if (is_valid_ether_addr(addr)){
		for (i=0;i<6;i++)
			mrv_reg_write8(priv, MRV_REG_BSSID, addr[i]);
		
		reg &= ~(2);
	}else{
		reg |= (2);
	}

	mrv_reg_write32(priv, MRV_REG_RX_MODE, reg);

}

static int mrv_read_config(struct mrv_priv *priv)
{
	struct mrv_cmd_hdr  *cmd;
	struct mrv_cmd_get_config *data;
	int i;

	cmd = mrv_cmd(priv, MRV_CMD_GET_CONFIG,
		MRV_CMD_GET_CONFIG_SZ, 0 ,1);
	
	data = (struct mrv_cmd_get_config *)MRV_CMD_PAYLOAD(cmd);
	data->Magic = cpu_to_le32(priv->magic_dma);
	
	for(i=0;i<MRV_CMD_GET_CONFIG_SZ;i++);
		((u8*)data)[i] = 0;
	
	for(i=0;i<6;i++)
		data->MacAdr[i] = 0xff;

	if(mrv_send_cmd_sync(priv)){
		printk(KERN_ERR"Could not read CFG\n");
		return -1;
	}

	priv->AntNum = le16_to_cpu(data->AntNum);
	priv->FwVer = le32_to_cpu(data->FwVer);
	priv->HwVer = le16_to_cpu(data->HwVer);
	priv->TxRingNum = le16_to_cpu(data->TxRingNum);
	priv->RegionCode = le16_to_cpu(data->RegionCode);
	priv->TxHwDma0 = le32_to_cpu(data->TxRing0) & 0xffff;
	priv->TxHwDma1 = le32_to_cpu(data->TxRing1) & 0xffff;
	priv->TxHwDma2 = le32_to_cpu(data->TxRing2) & 0xffff;
	priv->TxHwDma3 = le32_to_cpu(data->TxRing3) & 0xffff;
	priv->RxHwDmaR = le32_to_cpu(data->RxRingR) & 0xffff;
	priv->RxHwDmaW = le32_to_cpu(data->RxRingW) & 0xffff;

	SET_IEEE80211_PERM_ADDR(priv->dev, data->MacAdr);
	printk(KERN_INFO"Card mac address is "MAC_FMT"\n", 
			MAC_ARG(priv->dev->wiphy->perm_addr));
	printk(KERN_INFO"Fw V %x, Hw V %x, Anennas: %x, Region code %d\n", 
			priv->FwVer, priv->HwVer, priv->AntNum, priv->RegionCode);
	return 0;
}

static int mrv_alloc_rxring(struct mrv_priv *priv)
{
	struct mrv_rx_descriptor *v;///,*_v;
	dma_addr_t p;///,_p;
	int i;
///	_v = (struct mrv_rx_descriptor*) pci_alloc_consistent(priv->pdev, sizeof(struct mrv_rx_descriptor), &_p);
	for(i=0;i<MRV_RX_DESC_NUM;i++){
		v = (struct mrv_rx_descriptor*) pci_alloc_consistent(priv->pdev, sizeof(struct mrv_rx_descriptor), &p);
///		v = _v + i; p= _p + i*sizeof(struct mrv_rx_descriptor );
		if( v == NULL){
			printk(KERN_ERR"Can't allocate RX descriptor memory\n");
			/*FIXME dealloc what we done */
			return -ENOMEM;
		}
		
		v->drvowner = 0;
		v->status = cpu_to_le16(1);

		priv->rx_ring_dma[i] = p;
		priv->rx_ring_virt[i] = v;
		priv->rx_skb[i] = dev_alloc_skb(MRV_RX_PAYLOAD_LEN);
		
		if( priv->rx_skb[i] == NULL){
			printk(KERN_ERR"Can't allocate RX skb memory\n");
			/*FIXME dealloc what we done */
			return -ENOMEM;
		}
		
		priv->rx_ring_virt[i]->dma_data = cpu_to_le32(
				pci_map_single(priv->pdev,priv->rx_skb[i]->data,
				MRV_RX_PAYLOAD_LEN, PCI_DMA_FROMDEVICE));
		
		if(i != 0)
			priv->rx_ring_virt[i-1]->dma_next = cpu_to_le32(p);
	}
	priv->rx_ring_virt[i-1]->dma_next = cpu_to_le32(priv->rx_ring_dma[0]);
	priv->rx_idx = 0;
	
	return 0;
}

static int mrv_alloc_txring(struct mrv_priv *priv)
{
	struct mrv_tx_descriptor *v;///,*_v;
	dma_addr_t p;///,_p;
	int i;

	for(i=0;i<MRV_TX_DESC_NUM;i++){
		v = (struct mrv_tx_descriptor*) pci_alloc_consistent(priv->pdev, sizeof(struct mrv_tx_descriptor), &p);
///		v = _v + i; p= _p + i*sizeof(struct mrv_tx_descriptor );
		if( v == NULL){
			printk(KERN_ERR"Can't allocate TX descriptor memory\n");
			/*FIXME dealloc what we done */
			return -ENOMEM;
		}
		
		priv->tx_ring_dma[i] = p;
		priv->tx_ring_virt[i] = v;
		
		
		if(i != 0)
			priv->tx_ring_virt[i-1]->dma_next = cpu_to_le32(p);
	}
	priv->tx_ring_virt[i-1]->dma_next = cpu_to_le32(priv->tx_ring_dma[0]);
	priv->tx_r_idx = 0;
	priv->tx_w_idx = 0;
	priv->tx_pending = 0;

	return 0;
}

static void mrv_set_dma_rings(struct mrv_priv *priv)
{
	mrv_mem_write32(priv, priv->RxHwDmaR, priv->rx_ring_dma[0]);
	mrv_mem_write32(priv, priv->RxHwDmaW, priv->rx_ring_dma[0]);
	mrv_mem_write32(priv, priv->TxHwDma0, priv->tx_ring_dma[0]);
}

static int mrv_stop(struct ieee80211_hw *dev)
{
	struct mrv_priv *priv = dev->priv;	
	priv->up = 0;
	tasklet_disable(&priv->mrv_rx_tasklet);
	tasklet_disable(&priv->mrv_tx_cb_tasklet);
	mrv_mem_write32(priv, priv->TxHwDma0, 0);
	mrv_mem_write32(priv, priv->RxHwDmaR, 0);
	mrv_mem_write32(priv, priv->RxHwDmaW, 0);
	mrv_reg_mb(priv);
	msleep(100); /* random */
	tasklet_kill(&priv->mrv_rx_tasklet);
	tasklet_kill(&priv->mrv_tx_cb_tasklet);
	mrv_free_rx_ring(priv);
	mrv_free_tx_ring(priv);
	return 0;
}

static int mrv_open(struct ieee80211_hw *dev)
{
	struct mrv_priv *priv = dev->priv;
	priv->up = 1;
	tasklet_enable(&priv->mrv_rx_tasklet);
	tasklet_enable(&priv->mrv_tx_cb_tasklet);
	mrv_alloc_rxring(priv);
	mrv_alloc_txring(priv);
	mrv_set_dma_rings(priv);
	/* Suspect that all commands must be sent AFTER set radio. Including the
	 * get_hw_spec. But radio should be already enabled by default. Just make
	 * sure we use AUTO preamble here 
	*/
	mrv_set_radio(priv, 1, MRV_PREAMBLE_AUTO);
	mrv_set_antenna(priv, MRV_ANTENNA_RXA, 0xffff);
	mrv_set_antenna(priv, MRV_ANTENNA_TXA, 0x2);
	if(priv->AntNum == 2){
		mrv_set_antenna(priv, MRV_ANTENNA_RXB, 0xffff);
		mrv_set_antenna(priv, MRV_ANTENNA_TXB, 0x2);
	}
	mrv_set_txpower(priv, 10);

	return 0;
}	

static void mrv_set_promisc(struct mrv_priv *priv, int enable)
{
	u32 reg;
	reg = mrv_reg_read32(priv, MRV_REG_RX_MODE);
	if(enable)
		reg |= 1;
	else
		reg &= ~1;
	mrv_reg_write32(priv, MRV_REG_RX_MODE, reg);
}

static int mrv_config_interface(struct ieee80211_hw *dev, int if_id,
				    struct ieee80211_if_conf *conf)
{
	struct mrv_priv *priv = dev->priv;
	int err = 0;

	memcpy(priv->bssid, conf->bssid, ETH_ALEN);
	mrv_set_bssid(priv, priv->bssid);
	
	return err;
}

static int mrv_config(struct ieee80211_hw *dev, struct ieee80211_conf *conf)
{
	struct mrv_priv *priv = dev->priv;
	int err;

	err = mrv_set_channel(priv, conf->channel);
	if(err){
		printk(KERN_ERR"Failed to set channel\n");
		return -1;
	}
	
	/* TODO in monitor mode make sure it use IEEE80211G mode */
	if(conf->phymode == MODE_IEEE80211G) 
		err = mrv_set_allowed_rates(priv, 1);
	else
		err =mrv_set_allowed_rates(priv, 0);

	if(err){
		printk(KERN_ERR"Failed to set rates\n");
		return -1;
	}

	err = mrv_set_slot(priv, conf->flags & IEEE80211_CONF_SHORT_SLOT_TIME);
	if(err){
		printk(KERN_ERR"Failed to set slot\n");
		return -1;
	}
	/*we must set again bssid filter after changing ch ?*/
	mrv_set_bssid(priv,priv->bssid);

	/* we must set the promisc mode *after* changing ch */
	mrv_set_promisc(priv,priv->mode == IEEE80211_IF_TYPE_MNTR);

	return 0;
}

static int mrv_set_rts_threshold(struct ieee80211_hw *dev, u32 val)
{
	struct mrv_cmd_hdr  *cmd;
	struct mrv_cmd_set_rts_threshold *data;
	struct mrv_priv *priv = dev->priv;

	cmd = mrv_cmd(priv, MRV_CMD_SET_RTS_THRESHOLD,
		MRV_CMD_SET_RTS_THRESHOLD_SZ, 0 ,1);
	
	data = (struct mrv_cmd_set_rts_threshold *)MRV_CMD_PAYLOAD(cmd);

	data->threshold = cpu_to_le32(val);

	if(mrv_send_cmd_sync(priv)){
		printk(KERN_ERR"Could not set rts threshold\n");
		return -1;
	}
	
	return 0;
}

static void mrv_remove_interface(struct ieee80211_hw *dev,
				     struct ieee80211_if_init_conf *conf)
{
	struct mrv_priv *priv = dev->priv;
	priv->mode = IEEE80211_IF_TYPE_MGMT;
}

static int mrv_add_interface(struct ieee80211_hw *dev,
				 struct ieee80211_if_init_conf *conf)
{
	struct mrv_priv *priv = dev->priv;
	u8 bssid[] = {0,0,0,0,0,0};
	
	/* NOTE: using IEEE80211_IF_TYPE_MGMT to indicate no mode selected */
	if (priv->mode != IEEE80211_IF_TYPE_MGMT)
		return -1;

	memcpy(priv->bssid, bssid, ETH_ALEN);
	priv->mode = conf->type;
	mrv_set_mac_addr(priv, conf->mac_addr);

	return 0;
}

static void mrv_set_d80211_stuff(struct mrv_priv *priv)
{
	memcpy(priv->channels, mrv_channels, sizeof(mrv_channels));
	memcpy(priv->rates, mrv_rates, sizeof(mrv_rates));
	priv->modes[0].mode = MODE_IEEE80211G;
	priv->modes[0].num_rates = ARRAY_SIZE(mrv_rates);
	priv->modes[0].rates = priv->rates;
	priv->modes[0].num_channels = ARRAY_SIZE(mrv_channels);
	priv->modes[0].channels = priv->channels;
	priv->modes[1].mode = MODE_IEEE80211B;
	priv->modes[1].num_rates = 4;
	priv->modes[1].rates = priv->rates;
	priv->modes[1].num_channels = ARRAY_SIZE(mrv_channels);
	priv->modes[1].channels = priv->channels;
	priv->dev->flags = IEEE80211_HW_HOST_BROADCAST_PS_BUFFERING |
		     IEEE80211_HW_WEP_INCLUDE_IV |
		     IEEE80211_HW_DATA_NULLFUNC_ACK;
	priv->dev->extra_tx_headroom = 8; /* 6 bytes addr4, 2 bytes FW len */
	priv->dev->queues = 1;
	priv->dev->max_rssi = 100; // just to avoid dividing by zero 
//	priv->dev->channel_change_time = 100000;
	priv->mode = IEEE80211_IF_TYPE_MGMT;
}

static struct ieee80211_ops mrv_ops = {
	.tx				= mrv_tx,
	.open			= mrv_open,
	.stop			= mrv_stop,
	.add_interface	= mrv_add_interface,
	.remove_interface	= mrv_remove_interface,
	.config			= mrv_config,
	.config_interface	= mrv_config_interface,
	.set_rts_threshold	= mrv_set_rts_threshold,

};

static int __devinit mrv_init_one (struct pci_dev *pdev,
				     const struct pci_device_id *ent)
{
	struct ieee80211_hw *dev = NULL;
	struct mrv_priv *priv;
	unsigned long bar1_start, bar1_len, bar1_flags;
	unsigned long bar2_start, bar2_len, bar2_flags;
	void *bar1_addr=NULL, *bar2_addr=NULL;
	int err;

	pci_enable_device(pdev);
	
	bar1_start = pci_resource_start(pdev, 0);
	bar1_len = pci_resource_len(pdev, 0);
	bar1_flags = pci_resource_flags(pdev, 0);

	if (!(bar1_flags & IORESOURCE_MEM)) {
		printk(KERN_ERR"bar1 resource type is not memory\n");
		err = -ENODEV;
		goto fail_disable_device;
	}

	bar2_start = pci_resource_start(pdev, 1);
	bar2_len = pci_resource_len(pdev, 1);
	bar2_flags = pci_resource_flags(pdev, 1);

	if (!(bar2_flags & IORESOURCE_MEM)) {
		printk(KERN_ERR"bar1 resource type is not memory\n");
		err = -ENODEV;
		goto fail_disable_device;
	}

	err = pci_set_dma_mask(pdev, DMA_32BIT_MASK);
	if (err) {
		printk(KERN_ERR"No usable DMA configuration.\n");
		goto fail_disable_device;
	}

	bar1_addr = ioremap_nocache(bar1_start, bar1_len);
	if (bar1_addr == NULL) {
		printk(KERN_ERR"Error calling ioremap for bar1.\n");
		err = -EIO;
		goto fail_disable_device;
	}

	bar2_addr = ioremap_nocache(bar2_start, bar2_len);
	if (bar2_addr == NULL) {
		printk(KERN_ERR"Error calling ioremap for bar2.\n");
		err = -EIO;
		goto fail_iounmap;
	}

        err = pci_request_regions(pdev, DRV_NAME);
	if (err) {
		printk(KERN_ERR"Error calling pci_request_regions.\n");
		goto fail_iounmap;
	}
	pci_set_master(pdev);
	pci_set_mwi(pdev);

	
	dev = ieee80211_alloc_hw(sizeof(*priv), &mrv_ops);
	if (!dev) {
		printk(KERN_ERR "%s : can't allocate d80211 \n", pci_name(pdev));
		err = -ENOMEM;
		goto fail_release_regions;
	}

	priv = dev->priv;
	priv->pdev = pdev;
	priv->dev = dev;

	SET_MODULE_OWNER(dev);

	SET_IEEE80211_DEV(dev, &pdev->dev);
	pci_set_drvdata(pdev, dev);

	priv->bar1 = bar1_addr;
	priv->bar1_phys = bar1_start;
	priv->bar2 = bar2_addr;
	priv->bar2_phys = bar2_start;	
	
	sema_init(&priv->cmd_sem,1);

	tasklet_init(&priv->mrv_rx_tasklet,
		     (void(*)(unsigned long)) mrv_rx_tasklet,
		     (unsigned long)priv);

	tasklet_init(&priv->mrv_tx_cb_tasklet,
		     (void(*)(unsigned long)) mrv_tx_cb_tasklet,
		     (unsigned long)priv);
	
	tasklet_disable(&priv->mrv_rx_tasklet);
	tasklet_disable(&priv->mrv_tx_cb_tasklet);
	
	init_waitqueue_head(&priv->wait_command_queue);
	
	spin_lock_init(&priv->irq_th_lock);
	spin_lock_init(&priv->tx_lock);
	
	priv->fw_load = 0;
	priv->up = 0;	

	mrv_init_cmd(priv);
	
	err = mrv_hw_start(priv);
	if(err)
		goto fail_deinit_hw;

	mrv_read_config(priv);

	mrv_set_d80211_stuff(priv);
	
	err = ieee80211_register_hwmode(dev, &priv->modes[0]);
	if (err) {
		printk(KERN_ERR"%s can't register hwmode\n", pci_name(pdev));
		goto fail_stop_hw;
	}
	
	err = ieee80211_register_hwmode(dev, &priv->modes[1]);
	if (err) {
		printk(KERN_ERR"%s can't register hwmode\n", pci_name(pdev));
		goto fail_stop_hw;
	}
	
	err = ieee80211_register_hw(dev);
	if(err){
		printk(KERN_ERR"%s can't register HW\n", pci_name(pdev));
		goto fail_stop_hw;
	}

	return 0;
fail_stop_hw:
	mrv_hw_halt(priv);
fail_deinit_hw:
	mrv_deinit_cmd(priv);
	pci_set_drvdata(pdev, NULL);
	ieee80211_free_hw(dev);
fail_release_regions:
	pci_release_regions(pdev);
fail_iounmap:
	if (bar1_addr)
		iounmap(bar1_addr);
	if (bar2_addr)
		iounmap(bar2_addr);
fail_disable_device:
	pci_disable_device(pdev);
	pci_set_drvdata(pdev, NULL);

	return err;
}

static void __devexit mrv_remove_one (struct pci_dev *pdev)
{
	struct ieee80211_hw *dev = pci_get_drvdata(pdev);
	struct mrv_priv *priv;	
	
	if (dev) {
		priv = dev->priv;
		mrv_hw_halt(priv);
		
		ieee80211_unregister_hw(dev);
		mrv_deinit_cmd(priv);

		if (priv->bar1)
			iounmap(priv->bar1);
		if (priv->bar2)
			iounmap(priv->bar2);

		pci_release_regions(pdev);
		pci_disable_device(pdev);

		pci_set_drvdata (pdev, NULL);
		
		ieee80211_free_hw(dev);
	 }
}


static struct pci_driver mrv_pci_driver = {
	.name		= DRV_NAME,
	.id_table		= mrv_pci_tbl,
	.probe		= mrv_init_one,
	.remove		= __devexit_p(mrv_remove_one),
};


static int __init mrv_init_module (void)
{
	printk(KERN_INFO"%s \n", DRV_DESCRIPTION);
	printk(KERN_INFO"%s\n", DRV_COPYRIGHT);

	return pci_register_driver(&mrv_pci_driver);
}


static void __exit mrv_cleanup_module (void)
{
	pci_unregister_driver(&mrv_pci_driver);
}

module_init(mrv_init_module);
module_exit(mrv_cleanup_module);

